@@ -39,6 +39,7 @@ static gchar *separator;
 static gboolean print_all_columns = FALSE;
 static gint *print_columns = NULL;
 static gint *hide_columns = NULL;
+static GIOChannel *channel;
 
 static int *zenity_tree_extract_column_indexes (char *indexes, gint n_columns);
 static gboolean zenity_tree_column_is_hidden (gint column_index);
@@ -125,11 +126,17 @@ zenity_tree_handle_stdin (GIOChannel  *channel,
       gint status;
 
       do {
-        status = g_io_channel_read_line_string (channel, string, NULL, &error);
+        if (channel->is_readable == TRUE)
+          status = g_io_channel_read_line_string (channel, string, NULL, &error);
+        else
+          return FALSE;
 
         while (gtk_events_pending ())
           gtk_main_iteration ();
 
+        // TODO: Find a better way to avoid 100% cpu utilization
+        g_usleep(10000);
+
       } while (status == G_IO_STATUS_AGAIN);
 
       if (status != G_IO_STATUS_NORMAL) {
@@ -180,7 +187,7 @@ zenity_tree_handle_stdin (GIOChannel  *channel,
     g_string_free (string, TRUE);
   }
 
-  if (condition != G_IO_IN) {
+  if ((condition != G_IO_IN) && (condition != G_IO_IN + G_IO_HUP)) {
     g_io_channel_shutdown (channel, TRUE, NULL);
     return FALSE;
   }
@@ -193,8 +200,6 @@ zenity_tree_fill_entries_from_stdin (GtkTreeView  *tree_view,
                                      gboolean      toggles,
                                      gboolean      editable)
 {
-  GIOChannel *channel;
-
   g_object_set_data (G_OBJECT (tree_view), "n_columns", GINT_TO_POINTER (n_columns));
   g_object_set_data (G_OBJECT (tree_view), "toggles", GINT_TO_POINTER (toggles));
   g_object_set_data (G_OBJECT (tree_view), "editable", GINT_TO_POINTER (editable)); 
@@ -282,6 +287,7 @@ void
 zenity_tree (ZenityData *data, ZenityTreeData *tree_data)
 {
   GtkWidget *dialog;
+  GtkWidget *button;
   GObject *tree_view;
   GObject *text;
   GtkTreeViewColumn *column;
@@ -339,6 +345,20 @@ zenity_tree (ZenityData *data, ZenityTreeData *tree_data)
   if (data->dialog_title)
     gtk_window_set_title (GTK_WINDOW (dialog), data->dialog_title);
 
+  if (data->ok_label) {
+    button = GTK_WIDGET (gtk_builder_get_object (builder, "zenity_tree_ok_button"));
+    gtk_button_set_label (GTK_BUTTON (button), data->ok_label);
+    gtk_button_set_image (GTK_BUTTON (button),
+                          gtk_image_new_from_stock (GTK_STOCK_OK, GTK_ICON_SIZE_BUTTON));
+  }
+
+  if (data->cancel_label) {
+    button = GTK_WIDGET (gtk_builder_get_object (builder, "zenity_tree_cancel_button"));
+    gtk_button_set_label (GTK_BUTTON (button), data->cancel_label);
+    gtk_button_set_image (GTK_BUTTON (button),
+                          gtk_image_new_from_stock (GTK_STOCK_CANCEL, GTK_ICON_SIZE_BUTTON));
+  }
+
   text = gtk_builder_get_object (builder, "zenity_tree_text");
                                                                                 
   if (tree_data->dialog_text)
@@ -499,7 +519,7 @@ zenity_tree (ZenityData *data, ZenityTreeData *tree_data)
   zenity_util_show_dialog (dialog);
 
   if(data->timeout_delay > 0) {
-    g_timeout_add_seconds (data->timeout_delay, (GSourceFunc) zenity_util_timeout_handle, NULL);
+    g_timeout_add_seconds (data->timeout_delay, (GSourceFunc) zenity_util_timeout_handle, dialog);
   }
 
   gtk_main ();
@@ -613,7 +633,7 @@ zenity_tree_dialog_response (GtkWidget *widget, int response, gpointer data)
                                              GTK_TREE_VIEW (tree_view));
       }
       zenity_tree_dialog_output ();
-      zen_data->exit_code = zenity_util_return_exit_code (ZENITY_OK);
+      zenity_util_exit_code_with_data(ZENITY_OK, zen_data);
       break;
 
     case GTK_RESPONSE_CANCEL:
@@ -625,6 +645,9 @@ zenity_tree_dialog_response (GtkWidget *widget, int response, gpointer data)
       zen_data->exit_code = zenity_util_return_exit_code (ZENITY_ESC);
       break;
   }
+  if (channel != NULL)
+    g_io_channel_shutdown (channel, TRUE, NULL);
+
   gtk_main_quit ();
 }
 
